# (c) Copyright 2013 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os.path
from unittest import mock

import ddt

from os_brick.initiator import linuxfc
from os_brick.tests import base


@ddt.ddt
class LinuxFCTestCase(base.TestCase):

    def setUp(self):
        super(LinuxFCTestCase, self).setUp()
        self.cmds = []

        self.mock_object(os.path, 'exists', return_value=True)
        self.mock_object(os.path, 'isdir', return_value=True)
        self.lfc = linuxfc.LinuxFibreChannel(None, execute=self.fake_execute)

    def fake_execute(self, *cmd, **kwargs):
        self.cmds.append(" ".join(cmd))
        return "", None

    @staticmethod
    def __get_rescan_info(zone_manager=False):

        connection_properties = {
            'initiator_target_map': {'50014380186af83c': ['514f0c50023f6c00'],
                                     '50014380186af83e': ['514f0c50023f6c01']},
            'initiator_target_lun_map': {
                '50014380186af83c': [('514f0c50023f6c00', 1)],
                '50014380186af83e': [('514f0c50023f6c01', 1)]
            },
            'target_discovered': False,
            'target_lun': 1,
            'target_wwn': ['514f0c50023f6c00', '514f0c50023f6c01'],
            'targets': [
                ('514f0c50023f6c00', 1),
                ('514f0c50023f6c01', 1),
            ]
        }

        hbas = [
            {'device_path': ('/sys/devices/pci0000:00/0000:00:02.0/'
                             '0000:04:00.0/host6/fc_host/host6'),
             'host_device': 'host6',
             'node_name': '50014380186af83d',
             'port_name': '50014380186af83c'},
            {'device_path': ('/sys/devices/pci0000:00/0000:00:02.0/'
                             '0000:04:00.1/host7/fc_host/host7'),
             'host_device': 'host7',
             'node_name': '50014380186af83f',
             'port_name': '50014380186af83e'},
        ]
        if not zone_manager:
            del connection_properties['initiator_target_map']
            del connection_properties['initiator_target_lun_map']
        return hbas, connection_properties

    @staticmethod
    def _get_expected_info(wwpns=["514f0c50023f6c00", "514f0c50023f6c01"],
                           targets=1, remote_scan=False):

        execute_results = []
        expected_cmds = []
        for i in range(0, targets):
            expected_cmds += [
                mock.call(f'grep -Gil "{wwpns[i]}" '
                          '/sys/class/fc_transport/target6:*/port_name',
                          shell=True)
            ]
            if remote_scan:
                execute_results += [
                    # We can only perform remote ports scan if the
                    # fc_transport path returns empty output
                    ('', ''),
                    # This is returned from the fc_remote_ports path
                    ('/sys/class/fc_remote_ports/'
                     f'rport-6:0-{i+1}/port_name\n', ''),  # noqa: E226
                ]
                expected_cmds += [
                    mock.call(f'grep -Gil "{wwpns[i]}" '
                              '/sys/class/fc_remote_ports/rport-6:*/port_name',
                              shell=True),
                ]
            else:
                execute_results += [
                    ('/sys/class/fc_transport/'
                     f'target6:0:{i+1}/port_name\n', '')  # noqa: E226
                ]

        return execute_results, expected_cmds

    @mock.patch('builtins.open')
    @ddt.data(True, False)
    def test__get_hba_channel_scsi_target_lun_single_wwpn(
            self, remote_scan, mock_open):
        execute_results, expected_cmds = self._get_expected_info(
            remote_scan=remote_scan)
        if remote_scan:
            mock_open = mock_open.return_value.__enter__.return_value
            mock_open.read.return_value = ('1\n')
        hbas, con_props = self.__get_rescan_info()
        con_props['target_wwn'] = con_props['target_wwn'][0]
        con_props['targets'] = con_props['targets'][0:1]
        with mock.patch.object(self.lfc, '_execute',
                               side_effect=execute_results) as execute_mock:
            res = self.lfc._get_hba_channel_scsi_target_lun(hbas[0], con_props)
            execute_mock.assert_has_calls(expected_cmds)
        expected = ([['0', '1', 1]], set())
        self.assertEqual(expected, res)

    @mock.patch('builtins.open')
    @ddt.data(True, False)
    def test__get_hba_channel_scsi_target_lun_with_initiator_target_map(
            self, remote_scan, mock_open):
        execute_results, expected_cmds = self._get_expected_info(
            wwpns=["514f0c50023f6c01"])
        if remote_scan:
            mock_open = mock_open.return_value.__enter__.return_value
            mock_open.read.return_value = ('1\n')
        hbas, con_props = self.__get_rescan_info(zone_manager=True)
        con_props['target_wwn'] = con_props['target_wwn'][0]
        con_props['targets'] = con_props['targets'][0:1]
        hbas[0]['port_name'] = '50014380186af83e'
        with mock.patch.object(self.lfc, '_execute',
                               side_effect=execute_results) as execute_mock:
            res = self.lfc._get_hba_channel_scsi_target_lun(hbas[0], con_props)
            execute_mock.assert_has_calls(expected_cmds)
        expected = ([['0', '1', 1]], set())
        self.assertEqual(expected, res)

    @mock.patch('builtins.open')
    @ddt.data(True, False)
    def test__get_hba_channel_scsi_target_lun_with_initiator_target_map_none(
            self, remote_scan, mock_open):
        execute_results, expected_cmds = self._get_expected_info()
        if remote_scan:
            mock_open = mock_open.return_value.__enter__.return_value
            mock_open.read.return_value = ('1\n')
        hbas, con_props = self.__get_rescan_info()
        con_props['target_wwn'] = con_props['target_wwn'][0]
        con_props['targets'] = con_props['targets'][0:1]
        con_props['initiator_target_map'] = None
        hbas[0]['port_name'] = '50014380186af83e'
        with mock.patch.object(self.lfc, '_execute',
                               side_effect=execute_results) as execute_mock:
            res = self.lfc._get_hba_channel_scsi_target_lun(hbas[0], con_props)
            execute_mock.assert_has_calls(expected_cmds)
        expected = ([['0', '1', 1]], set())
        self.assertEqual(expected, res)

    @mock.patch('builtins.open')
    @ddt.data(True, False)
    def test__get_hba_channel_scsi_target_lun_multiple_wwpn(
            self, remote_scan, mock_open):
        execute_results, expected_cmds = self._get_expected_info(targets=2)
        if remote_scan:
            mock_open = mock_open.return_value.__enter__.return_value
            mock_open.read.return_value = ('1\n')
        hbas, con_props = self.__get_rescan_info()
        with mock.patch.object(self.lfc, '_execute',
                               side_effect=execute_results) as execute_mock:
            res = self.lfc._get_hba_channel_scsi_target_lun(hbas[0], con_props)
            execute_mock.assert_has_calls(expected_cmds)

        expected = ([['0', '1', 1], ['0', '2', 1]], set())
        self.assertEqual(expected, res)

    @mock.patch('builtins.open')
    @ddt.data(True, False)
    def test__get_hba_channel_scsi_target_lun_multiple_wwpn_and_luns(
            self, remote_scan, mock_open):
        execute_results, expected_cmds = self._get_expected_info(targets=2)
        if remote_scan:
            mock_open = mock_open.return_value.__enter__.return_value
            mock_open.read.return_value = ('1\n')
        hbas, con_props = self.__get_rescan_info()
        con_props['target_lun'] = [1, 7]
        con_props['targets'] = [
            ('514f0c50023f6c00', 1),
            ('514f0c50023f6c01', 7),
        ]
        with mock.patch.object(self.lfc, '_execute',
                               side_effect=execute_results) as execute_mock:
            res = self.lfc._get_hba_channel_scsi_target_lun(hbas[0], con_props)
            execute_mock.assert_has_calls(expected_cmds)

        expected = ([['0', '1', 1], ['0', '2', 7]], set())
        self.assertEqual(expected, res)

    @mock.patch('builtins.open')
    @ddt.data(True, False)
    def test__get_hba_channel_scsi_target_lun_zone_manager(
            self, remote_scan, mock_open):
        execute_results, expected_cmds = self._get_expected_info()
        if remote_scan:
            mock_open = mock_open.return_value.__enter__.return_value
            mock_open.read.return_value = ('1\n')
        hbas, con_props = self.__get_rescan_info(zone_manager=True)
        with mock.patch.object(self.lfc, '_execute',
                               side_effect=execute_results) as execute_mock:
            res = self.lfc._get_hba_channel_scsi_target_lun(hbas[0], con_props)
            execute_mock.assert_has_calls(expected_cmds)
        expected = ([['0', '1', 1]], set())
        self.assertEqual(expected, res)

    def test__get_hba_channel_scsi_target_lun_both_paths_not_found(self):
        _, expected_cmds = self._get_expected_info()
        hbas, con_props = self.__get_rescan_info(zone_manager=True)
        with mock.patch.object(self.lfc, '_execute',
                               return_value=('', '')) as execute_mock:
            res = self.lfc._get_hba_channel_scsi_target_lun(hbas[0], con_props)
            execute_mock.assert_has_calls(expected_cmds)
        self.assertEqual(([], {1}), res)

    def test__get_hba_channel_scsi_target_lun_exception(self):
        _, expected_cmds = self._get_expected_info()
        hbas, con_props = self.__get_rescan_info(zone_manager=True)
        with mock.patch.object(self.lfc, '_execute',
                               side_effect=Exception) as execute_mock:
            res = self.lfc._get_hba_channel_scsi_target_lun(hbas[0], con_props)
            execute_mock.assert_has_calls(expected_cmds)
        self.assertEqual(([], {1}), res)

    def test__get_hba_channel_scsi_target_lun_fc_transport_exception(self):
        execute_effects = [
            ('/sys/class/fc_transport/target6:0:1/port_name\n', ''),
            Exception()
        ]
        _, expected_cmds = self._get_expected_info()
        hbas, con_props = self.__get_rescan_info()

        with mock.patch.object(self.lfc, '_execute',
                               side_effect=execute_effects) as execute_mock:
            res = self.lfc._get_hba_channel_scsi_target_lun(hbas[0], con_props)
            execute_mock.assert_has_calls(expected_cmds)
        expected = ([['0', '1', 1]], {1})
        self.assertEqual(expected, res)

    @mock.patch('builtins.open')
    def test__get_hba_channel_scsi_target_lun_fc_remote_ports_exception(
            self, mock_open):
        execute_effects = [
            ('', ''),
            ('/sys/class/fc_remote_ports/rport-6:0-1/port_name\n', ''),
            ('', ''),
            Exception()
        ]
        mock_open = mock_open.return_value.__enter__.return_value
        mock_open.read.return_value = ('1\n')
        _, expected_cmds = self._get_expected_info()
        hbas, con_props = self.__get_rescan_info()

        with mock.patch.object(self.lfc, '_execute',
                               side_effect=execute_effects) as execute_mock:
            res = self.lfc._get_hba_channel_scsi_target_lun(hbas[0], con_props)
            execute_mock.assert_has_calls(expected_cmds)
        expected = ([['0', '1', 1]], {1})
        self.assertEqual(expected, res)

    @mock.patch('builtins.open')
    def test__get_hba_channel_scsi_target_open_oserror(
            self, mock_open):
        execute_effects, expected_cmds = self._get_expected_info(
            targets=2, remote_scan=True)
        mock_open = mock_open.return_value.__enter__.return_value
        mock_open.read.side_effect = ['1\n', OSError()]
        hbas, con_props = self.__get_rescan_info()

        with mock.patch.object(self.lfc, '_execute',
                               side_effect=execute_effects) as execute_mock:
            res = self.lfc._get_hba_channel_scsi_target_lun(hbas[0], con_props)
            execute_mock.assert_has_calls(expected_cmds)
        expected = ([['0', '1', 1]], {1})
        self.assertEqual(expected, res)

    def test__get_target_fc_transport_path(self):
        path = '/sys/class/fc_transport/target6:'
        execute_results = ('/sys/class/fc_transport/target6:0:1/port_name\n',
                           '')

        _, con_props = self.__get_rescan_info()
        with mock.patch.object(self.lfc, '_execute',
                               return_value=execute_results) as execute_mock:
            ctl = self.lfc._get_target_fc_transport_path(
                path, con_props['target_wwn'][0], 1)
            execute_mock.assert_called_once_with(
                'grep -Gil "514f0c50023f6c00" '
                '/sys/class/fc_transport/target6:*/port_name',
                shell=True)
        self.assertEqual(['0', '1', 1], ctl)

    @mock.patch('builtins.open')
    def test__get_target_fc_remote_ports_path(self, mock_open):
        path = '/sys/class/fc_remote_ports/rport-6:'
        execute_results = [
            ('/sys/class/fc_remote_ports/rport-6:0-1/port_name\n', ''),
            ('1\n', ''),
        ]
        scsi_target_path = (
            '/sys/class/fc_remote_ports/rport-6:0-1/scsi_target_id')
        mock_open.return_value.__enter__.return_value.read.return_value = (
            '1\n')
        hbas, con_props = self.__get_rescan_info()
        with mock.patch.object(self.lfc, '_execute',
                               side_effect=execute_results) as execute_mock:
            ctl = self.lfc._get_target_fc_remote_ports_path(
                path, con_props['target_wwn'][0], 1)
            expected_cmds = [
                mock.call(
                    'grep -Gil "514f0c50023f6c00" '
                    '/sys/class/fc_remote_ports/rport-6:*/port_name',
                    shell=True),
            ]
            execute_mock.assert_has_calls(expected_cmds)
            mock_open.assert_called_once_with(scsi_target_path)

        self.assertEqual(['0', '1', 1], ctl)

    def test_rescan_hosts_initiator_map(self):
        """Test FC rescan with initiator map and not every HBA connected."""
        get_chan_results = [([['2', '3', 1], ['4', '5', 1]], set()),
                            ([['6', '7', 1]], set())]

        hbas, con_props = self.__get_rescan_info(zone_manager=True)

        # This HBA is not in the initiator map, so we should not scan it or try
        # to get the channel and target
        hbas.append({'device_path': ('/sys/devices/pci0000:00/0000:00:02.0/'
                                     '0000:04:00.2/host8/fc_host/host8'),
                     'host_device': 'host8',
                     'node_name': '50014380186af83g',
                     'port_name': '50014380186af83h'})

        with mock.patch.object(self.lfc, '_execute',
                               return_value=None) as execute_mock, \
            mock.patch.object(self.lfc, '_get_hba_channel_scsi_target_lun',
                              side_effect=get_chan_results) as mock_get_chan:

            self.lfc.rescan_hosts(hbas, con_props)
            expected_commands = [
                mock.call('tee', '-a', '/sys/class/scsi_host/host6/scan',
                          process_input='2 3 1',
                          root_helper=None, run_as_root=True),
                mock.call('tee', '-a', '/sys/class/scsi_host/host6/scan',
                          process_input='4 5 1',
                          root_helper=None, run_as_root=True),
                mock.call('tee', '-a', '/sys/class/scsi_host/host7/scan',
                          process_input='6 7 1',
                          root_helper=None, run_as_root=True)]

            execute_mock.assert_has_calls(expected_commands)
            self.assertEqual(len(expected_commands), execute_mock.call_count)

            expected_calls = [mock.call(hbas[0], con_props),
                              mock.call(hbas[1], con_props)]
            mock_get_chan.assert_has_calls(expected_calls)

    @mock.patch.object(linuxfc.LinuxFibreChannel, 'lun_for_addressing')
    def test_rescan_hosts_single_wwnn(self, lun_addr_mock):
        """Test FC rescan with no initiator map and single WWNN for ports."""
        lun_addr_mock.return_value = 16640
        get_chan_results = [
            [[['2', '3', 256], ['4', '5', 256]], set()],
            [[['6', '7', 256]], set()],
            [[], {1}],
        ]

        hbas, con_props = self.__get_rescan_info(zone_manager=False)
        con_props['addressing_mode'] = 'SAM2'

        # This HBA is the one that is not included in the single WWNN.
        hbas.append({'device_path': ('/sys/devices/pci0000:00/0000:00:02.0/'
                                     '0000:04:00.2/host8/fc_host/host8'),
                     'host_device': 'host8',
                     'node_name': '50014380186af83g',
                     'port_name': '50014380186af83h'})

        with mock.patch.object(self.lfc, '_execute',
                               return_value=None) as execute_mock, \
            mock.patch.object(self.lfc, '_get_hba_channel_scsi_target_lun',
                              side_effect=get_chan_results) as mock_get_chan:

            self.lfc.rescan_hosts(hbas, con_props)
            expected_commands = [
                mock.call('tee', '-a', '/sys/class/scsi_host/host6/scan',
                          process_input='2 3 16640',
                          root_helper=None, run_as_root=True),
                mock.call('tee', '-a', '/sys/class/scsi_host/host6/scan',
                          process_input='4 5 16640',
                          root_helper=None, run_as_root=True),
                mock.call('tee', '-a', '/sys/class/scsi_host/host7/scan',
                          process_input='6 7 16640',
                          root_helper=None, run_as_root=True)]

            execute_mock.assert_has_calls(expected_commands)
            self.assertEqual(len(expected_commands), execute_mock.call_count)

            expected_calls = [mock.call(hbas[0], con_props),
                              mock.call(hbas[1], con_props)]
            mock_get_chan.assert_has_calls(expected_calls)
        lun_addr_mock.assert_has_calls([mock.call(256, 'SAM2')] * 3)

    def test_rescan_hosts_initiator_map_single_wwnn(self):
        """Test FC rescan with initiator map and single WWNN."""
        get_chan_results = [([['2', '3', 1], ['4', '5', 1]], set()),
                            ([], {1})]

        hbas, con_props = self.__get_rescan_info(zone_manager=True)

        with mock.patch.object(self.lfc, '_execute',
                               return_value=None) as execute_mock, \
            mock.patch.object(self.lfc, '_get_hba_channel_scsi_target_lun',
                              side_effect=get_chan_results) as mock_get_chan:

            self.lfc.rescan_hosts(hbas, con_props)
            expected_commands = [
                mock.call('tee', '-a', '/sys/class/scsi_host/host6/scan',
                          process_input='2 3 1',
                          root_helper=None, run_as_root=True),
                mock.call('tee', '-a', '/sys/class/scsi_host/host6/scan',
                          process_input='4 5 1',
                          root_helper=None, run_as_root=True)]

            execute_mock.assert_has_calls(expected_commands)
            self.assertEqual(len(expected_commands), execute_mock.call_count)

            expected_calls = [mock.call(hbas[0], con_props),
                              mock.call(hbas[1], con_props)]
            mock_get_chan.assert_has_calls(expected_calls)

    def test_rescan_hosts_port_not_found(self):
        """Test when we don't find the target ports."""
        get_chan_results = [([], {1}), ([], {1})]
        hbas, con_props = self.__get_rescan_info(zone_manager=True)
        # Remove the initiator map
        con_props.pop('initiator_target_map')
        con_props.pop('initiator_target_lun_map')
        with mock.patch.object(self.lfc, '_get_hba_channel_scsi_target_lun',
                               side_effect=get_chan_results), \
            mock.patch.object(self.lfc, '_execute',
                              side_effect=None) as execute_mock:

            self.lfc.rescan_hosts(hbas, con_props)

            expected_commands = [
                mock.call('tee', '-a', '/sys/class/scsi_host/host6/scan',
                          process_input='- - 1',
                          root_helper=None, run_as_root=True),
                mock.call('tee', '-a', '/sys/class/scsi_host/host7/scan',
                          process_input='- - 1',
                          root_helper=None, run_as_root=True)]
            execute_mock.assert_has_calls(expected_commands)
            self.assertEqual(len(expected_commands), execute_mock.call_count)

    def test_rescan_hosts_port_not_found_driver_disables_wildcards(self):
        """Test when we don't find the target ports but driver forces scan."""
        get_chan_results = [([], {1}), ([], {1})]
        hbas, con_props = self.__get_rescan_info(zone_manager=True)
        con_props['enable_wildcard_scan'] = False
        with mock.patch.object(self.lfc, '_get_hba_channel_scsi_target_lun',
                               side_effect=get_chan_results), \
            mock.patch.object(self.lfc, '_execute',
                              side_effect=None) as execute_mock:

            self.lfc.rescan_hosts(hbas, con_props)
            execute_mock.assert_not_called()

    @mock.patch('glob.glob', return_value=[])
    def test_get_fc_hbas_no_hbas(self, mock_glob):
        hbas = self.lfc.get_fc_hbas()
        self.assertListEqual([], hbas)
        mock_glob.assert_called_once_with('/sys/class/fc_host/*')

    @mock.patch('os.path.realpath')
    @mock.patch('glob.glob', return_value=['/sys/class/fc_host/host0',
                                           '/sys/class/fc_host/host2'])
    @mock.patch('builtins.open', side_effect=IOError)
    def test_get_fc_hbas_fail(self, mock_open, mock_glob, mock_path):
        hbas = self.lfc.get_fc_hbas()
        mock_glob.assert_called_once_with('/sys/class/fc_host/*')
        self.assertListEqual([], hbas)
        self.assertEqual(2, mock_open.call_count)
        mock_open.assert_has_calls(
            (mock.call('/sys/class/fc_host/host0/port_name', 'rt'),
             mock.call('/sys/class/fc_host/host2/port_name', 'rt'))
        )
        self.assertEqual(2, mock_path.call_count)
        mock_path.assert_has_calls(
            (mock.call('/sys/class/fc_host/host0'),
             mock.call('/sys/class/fc_host/host2'))
        )

    @mock.patch('os.path.realpath')
    @mock.patch('glob.glob', return_value=['/sys/class/fc_host/host0',
                                           '/sys/class/fc_host/host2'])
    @mock.patch('builtins.open')
    def test_get_fc_hbas(self, mock_open, mock_glob, mock_path):
        mock_open.return_value.__enter__.return_value.read.side_effect = [
            '0x50014380242b9750\n', '0x50014380242b9751\n', 'Online',
            '0x50014380242b9752\n', '0x50014380242b9753\n', 'Online',
        ]
        pci_path = '/sys/devices/pci0000:20/0000:20:03.0/0000:21:00.'
        host0_pci = f'{pci_path}0/host0/fc_host/host0'
        host2_pci = f'{pci_path}1/host2/fc_host/host2'
        mock_path.side_effect = [host0_pci, host2_pci]

        hbas = self.lfc.get_fc_hbas()

        expected = [
            {'ClassDevice': 'host0',
             'ClassDevicepath': host0_pci,
             'port_name': '0x50014380242b9750',
             'node_name': '0x50014380242b9751',
             'port_state': 'Online'},
            {'ClassDevice': 'host2',
             'ClassDevicepath': host2_pci,
             'port_name': '0x50014380242b9752',
             'node_name': '0x50014380242b9753',
             'port_state': 'Online'},
        ]
        self.assertListEqual(expected, hbas)
        mock_glob.assert_called_once_with('/sys/class/fc_host/*')
        self.assertEqual(6, mock_open.call_count)
        mock_open.assert_has_calls(
            (mock.call('/sys/class/fc_host/host0/port_name', 'rt'),
             mock.call('/sys/class/fc_host/host0/node_name', 'rt'),
             mock.call('/sys/class/fc_host/host0/port_state', 'rt'),
             mock.call('/sys/class/fc_host/host2/port_name', 'rt'),
             mock.call('/sys/class/fc_host/host2/node_name', 'rt'),
             mock.call('/sys/class/fc_host/host2/port_state', 'rt')),
            any_order=True,
        )
        self.assertEqual(2, mock_path.call_count)
        mock_path.assert_has_calls(
            (mock.call('/sys/class/fc_host/host0'),
             mock.call('/sys/class/fc_host/host2'))
        )

    def _set_get_fc_hbas(self):
        pci_path = '/sys/devices/pci0000:20/0000:20:03.0/0000:21:00.'
        host0_pci = f'{pci_path}0/host0/fc_host/host0'
        host2_pci = f'{pci_path}1/host2/fc_host/host2'
        return_value = [{'ClassDevice': 'host0',
                         'ClassDevicepath': host0_pci,
                         'port_name': '0x50014380242b9750',
                         'node_name': '0x50014380242b9751',
                         'port_state': 'Online'},
                        {'ClassDevice': 'host2',
                         'ClassDevicepath': host2_pci,
                         'port_name': '0x50014380242b9752',
                         'node_name': '0x50014380242b9753',
                         'port_state': 'Online'}]
        mocked = self.mock_object(linuxfc.LinuxFibreChannel, 'get_fc_hbas',
                                  return_value=return_value)
        return mocked

    def test_get_fc_hbas_info(self):
        mock_hbas = self._set_get_fc_hbas()
        hbas_info = self.lfc.get_fc_hbas_info()
        expected_info = [{'device_path': '/sys/devices/pci0000:20/'
                                         '0000:20:03.0/0000:21:00.0/'
                                         'host0/fc_host/host0',
                          'host_device': 'host0',
                          'node_name': '50014380242b9751',
                          'port_name': '50014380242b9750'},
                         {'device_path': '/sys/devices/pci0000:20/'
                                         '0000:20:03.0/0000:21:00.1/'
                                         'host2/fc_host/host2',
                          'host_device': 'host2',
                          'node_name': '50014380242b9753',
                          'port_name': '50014380242b9752'}, ]
        self.assertListEqual(expected_info, hbas_info)
        mock_hbas.assert_called_once_with()

    def test_get_fc_wwpns(self):
        self._set_get_fc_hbas()
        wwpns = self.lfc.get_fc_wwpns()
        expected_wwpns = ['50014380242b9750', '50014380242b9752']
        self.assertEqual(expected_wwpns, wwpns)

    def test_get_fc_wwnns(self):
        self._set_get_fc_hbas()
        wwnns = self.lfc.get_fc_wwnns()
        expected_wwnns = ['50014380242b9751', '50014380242b9753']
        self.assertEqual(expected_wwnns, wwnns)


class LinuxFCS390XTestCase(LinuxFCTestCase):

    def setUp(self):
        super(LinuxFCS390XTestCase, self).setUp()
        self.cmds = []
        self.lfc = linuxfc.LinuxFibreChannelS390X(None,
                                                  execute=self.fake_execute)

    @mock.patch.object(linuxfc.LinuxFibreChannel, 'get_fc_hbas')
    def test_get_fc_hbas_info(self, mock_hbas):
        host_pci = '/sys/devices/css0/0.0.02ea/0.0.3080/host0/fc_host/host0'
        mock_hbas.return_value = [{'ClassDevice': 'host0',
                                   'ClassDevicepath': host_pci,
                                   'port_name': '0xc05076ffe680a960',
                                   'node_name': '0x1234567898765432',
                                   'port_state': 'Online'}]
        hbas_info = self.lfc.get_fc_hbas_info()
        expected = [{'device_path': '/sys/devices/css0/0.0.02ea/'
                                    '0.0.3080/host0/fc_host/host0',
                     'host_device': 'host0',
                     'node_name': '1234567898765432',
                     'port_name': 'c05076ffe680a960'}]
        self.assertEqual(expected, hbas_info)

    @mock.patch.object(os.path, 'exists', return_value=False)
    def test_configure_scsi_device(self, mock_execute):
        device_number = "0.0.2319"
        target_wwn = "0x50014380242b9751"
        lun = 1
        self.lfc.configure_scsi_device(device_number, target_wwn, lun)
        expected_commands = [('tee -a /sys/bus/ccw/drivers/zfcp/0.0.2319/'
                             'port_rescan'),
                             ('tee -a /sys/bus/ccw/drivers/zfcp/0.0.2319/'
                                 '0x50014380242b9751/unit_add')]
        self.assertEqual(expected_commands, self.cmds)

    def test_deconfigure_scsi_device(self):
        device_number = "0.0.2319"
        target_wwn = "0x50014380242b9751"
        lun = 1
        self.lfc.deconfigure_scsi_device(device_number, target_wwn, lun)
        expected_commands = [('tee -a /sys/bus/ccw/drivers/zfcp/'
                              '0.0.2319/0x50014380242b9751/unit_remove')]
        self.assertEqual(expected_commands, self.cmds)



# Copyright (c) 2013 OpenStack Foundation
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Remote filesystem client utilities."""

import hashlib
import os
import re
import tempfile

from oslo_concurrency import processutils
from oslo_log import log as logging

from os_brick import exception
from os_brick import executor
from os_brick.i18n import _

LOG = logging.getLogger(__name__)


class RemoteFsClient(executor.Executor):

    def __init__(self, mount_type, root_helper,
                 execute=None, *args, **kwargs):
        super(RemoteFsClient, self).__init__(root_helper, execute=execute,
                                             *args, **kwargs)  # type: ignore

        mount_type_to_option_prefix = {
            'nfs': 'nfs',
            'cifs': 'smbfs',
            'glusterfs': 'glusterfs',
            'vzstorage': 'vzstorage',
            'quobyte': 'quobyte',
            'scality': 'scality'
        }

        if mount_type not in mount_type_to_option_prefix:
            raise exception.ProtocolNotSupported(protocol=mount_type)

        self._mount_type = mount_type
        option_prefix = mount_type_to_option_prefix[mount_type]

        self._mount_base: str
        self._mount_base = kwargs.get(option_prefix +
                                      '_mount_point_base')  # type: ignore
        if not self._mount_base:
            raise exception.InvalidParameterValue(
                err=_('%s_mount_point_base required') % option_prefix)

        self._mount_options = kwargs.get(option_prefix + '_mount_options')

        if mount_type == "nfs":
            self._check_nfs_options()

    def get_mount_base(self):
        return self._mount_base

    def _get_hash_str(self, base_str):
        """Return a string that represents hash of base_str (hex format)."""
        if isinstance(base_str, str):
            base_str = base_str.encode('utf-8')
        return hashlib.md5(base_str, usedforsecurity=False).hexdigest()

    def get_mount_point(self, device_name: str):
        """Get Mount Point.

        :param device_name: example 172.18.194.100:/var/nfs
        """
        return os.path.join(self._mount_base,
                            self._get_hash_str(device_name))

    def _read_mounts(self):
        """Returns a dict of mounts and their mountpoint

        Format reference:
        http://man7.org/linux/man-pages/man5/fstab.5.html
        """
        with open("/proc/mounts", "r") as mounts:
            # Remove empty lines and split lines by whitespace
            lines = [line.split() for line in mounts.read().splitlines()
                     if line.strip()]

            # Return {mountpoint: mountdevice}.  Fields 2nd and 1st as per
            # http://man7.org/linux/man-pages/man5/fstab.5.html
            return {line[1]: line[0] for line in lines if line[0] != '#'}

    def mount(self, share, flags=None):
        """Mount given share."""
        mount_path = self.get_mount_point(share)

        if mount_path in self._read_mounts():
            LOG.debug('Already mounted: %s', mount_path)
            return

        self._execute('mkdir', '-p', mount_path, check_exit_code=0)
        if self._mount_type == 'nfs':
            self._mount_nfs(share, mount_path, flags)
        else:
            self._do_mount(self._mount_type, share, mount_path,
                           self._mount_options, flags)

    def _do_mount(self, mount_type, share, mount_path, mount_options=None,
                  flags=None):
        """Mounts share based on the specified params."""
        mnt_cmd = ['mount', '-t', mount_type]
        if mount_options is not None:
            mnt_cmd.extend(['-o', mount_options])
        if flags is not None:
            mnt_cmd.extend(flags)
        mnt_cmd.extend([share, mount_path])

        try:
            self._execute(*mnt_cmd, root_helper=self._root_helper,
                          run_as_root=True, check_exit_code=0)
        except processutils.ProcessExecutionError as exc:
            if 'already mounted' in exc.stderr:
                LOG.debug("Already mounted: %s", share)

                # The error message can say "busy or already mounted" when the
                # share didn't actually mount, so look for it.
                if share in self._read_mounts():
                    return

            LOG.error("Failed to mount %(share)s, reason: %(reason)s",
                      {'share': share, 'reason': exc.stderr})
            raise

    def _mount_nfs(self, nfs_share, mount_path, flags=None):
        """Mount nfs share using present mount types."""
        mnt_errors = {}

        # This loop allows us to first try to mount with NFS 4.1 for pNFS
        # support but falls back to mount NFS 4 or NFS 3 if either the client
        # or server do not support it.
        for mnt_type in sorted(self._nfs_mount_type_opts.keys(), reverse=True):
            options = self._nfs_mount_type_opts[mnt_type]
            try:
                self._do_mount('nfs', nfs_share, mount_path, options, flags)
                LOG.debug('Mounted %(sh)s using %(mnt_type)s.',
                          {'sh': nfs_share, 'mnt_type': mnt_type})
                return
            except Exception as e:
                mnt_errors[mnt_type] = str(e)
                LOG.debug('Failed to do %s mount.', mnt_type)
        raise exception.BrickException(_("NFS mount failed for share %(sh)s. "
                                         "Error - %(error)s")
                                       % {'sh': nfs_share,
                                          'error': mnt_errors})

    def _check_nfs_options(self):
        """Checks and prepares nfs mount type options."""
        self._nfs_mount_type_opts = {'nfs': self._mount_options}
        nfs_vers_opt_patterns = ['^nfsvers', '^vers', r'^v[\d]']
        for opt in nfs_vers_opt_patterns:
            if self._option_exists(self._mount_options, opt):
                return

        # pNFS requires NFS 4.1. The mount.nfs4 utility does not automatically
        # negotiate 4.1 support, we have to ask for it by specifying two
        # options: vers=4 and minorversion=1.
        pnfs_opts = self._update_option(self._mount_options, 'vers', '4')
        pnfs_opts = self._update_option(pnfs_opts, 'minorversion', '1')
        self._nfs_mount_type_opts['pnfs'] = pnfs_opts

    def _option_exists(self, options, opt_pattern):
        """Checks if the option exists in nfs options and returns position."""
        options = [x.strip() for x in options.split(',')] if options else []
        pos = 0
        for opt in options:
            pos = pos + 1
            if re.match(opt_pattern, opt, flags=0):
                return pos
        return 0

    def _update_option(self, options, option, value=None):
        """Update option if exists else adds it and returns new options."""
        opts = [x.strip() for x in options.split(',')] if options else []
        pos = self._option_exists(options, option)
        if pos:
            opts.pop(pos - 1)
        opt = '%s=%s' % (option, value) if value else option
        opts.append(opt)
        return ",".join(opts) if len(opts) > 1 else opts[0]


class ScalityRemoteFsClient(RemoteFsClient):
    def __init__(self, mount_type, root_helper,
                 execute=None, *args, **kwargs):
        super(ScalityRemoteFsClient, self).__init__(
            mount_type, root_helper, execute=execute,
            *args, **kwargs)  # type: ignore
        self._mount_type = mount_type
        self._mount_base = kwargs.get(
            'scality_mount_point_base', "").rstrip('/')
        if not self._mount_base:
            raise exception.InvalidParameterValue(
                err=_('scality_mount_point_base required'))
        self._mount_options = None

    def get_mount_point(self, device_name):
        return os.path.join(self._mount_base,
                            device_name,
                            "00")

    def mount(self, share, flags=None):
        """Mount the Scality ScaleOut FS.

        The `share` argument is ignored because you can't mount several
        SOFS at the same type on a single server. But we want to keep the
        same method signature for class inheritance purpose.
        """
        if self._mount_base in self._read_mounts():
            LOG.debug('Already mounted: %s', self._mount_base)
            return
        self._execute('mkdir', '-p', self._mount_base, check_exit_code=0)
        super(ScalityRemoteFsClient, self)._do_mount(
            'sofs', '/etc/sfused.conf', self._mount_base)


class VZStorageRemoteFSClient(RemoteFsClient):
    def _vzstorage_write_mds_list(self, cluster_name, mdss):
        tmp_dir = tempfile.mkdtemp(prefix='vzstorage-')
        tmp_bs_path = os.path.join(tmp_dir, 'bs_list')
        with open(tmp_bs_path, 'w') as f:
            for mds in mdss:
                f.write(mds + "\n")

        conf_dir = os.path.join('/etc/pstorage/clusters', cluster_name)
        if os.path.exists(conf_dir):
            bs_path = os.path.join(conf_dir, 'bs_list')
            self._execute('cp', '-f', tmp_bs_path, bs_path,
                          root_helper=self._root_helper, run_as_root=True)
        else:
            self._execute('cp', '-rf', tmp_dir, conf_dir,
                          root_helper=self._root_helper, run_as_root=True)
        self._execute('chown', '-R', 'root:root', conf_dir,
                      root_helper=self._root_helper, run_as_root=True)

    def _do_mount(self, mount_type, vz_share, mount_path,
                  mount_options=None, flags=None):
        m = re.search(r"(?:(\S+):\/)?([a-zA-Z0-9_-]+)(?::(\S+))?", vz_share)
        if not m:
            msg = (_("Invalid Virtuozzo Storage share specification: %r."
                     "Must be: [MDS1[,MDS2],...:/]<CLUSTER NAME>[:PASSWORD].")
                   % vz_share)
            raise exception.BrickException(msg)

        mdss = m.group(1)
        cluster_name = m.group(2)
        passwd = m.group(3)

        if mdss:
            mdss = mdss.split(',')
            self._vzstorage_write_mds_list(cluster_name, mdss)

        if passwd:
            self._execute('pstorage', '-c', cluster_name, 'auth-node', '-P',
                          process_input=passwd,
                          root_helper=self._root_helper, run_as_root=True)

        mnt_cmd = ['pstorage-mount', '-c', cluster_name]
        if flags:
            mnt_cmd.extend(flags)
        mnt_cmd.extend([mount_path])

        self._execute(*mnt_cmd, root_helper=self._root_helper,
                      run_as_root=True, check_exit_code=0)



#    Copyright (c) 2015 - 2024 StorPool
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import json
import os
from unittest import mock


from os_brick import exception
from os_brick.initiator import storpool_utils
from os_brick.tests import base

STORPOOL_CONF_INI_NO_HOSTNAME = """\
SP_API_HTTP_HOST=127.0.0.1
SP_API_HTTP_PORT=81
SP_AUTH_TOKEN=1234567890123456789

[another-node]
SP_OURID=2
"""

STORPOOL_CONF_INI = STORPOOL_CONF_INI_NO_HOSTNAME + """\
[this-node]
SP_OURID=1
"""

STORPOOL_CONF_INI_SELECTOR = "full"

ANOTHER_CONF_INI = """\
SP_API_HTTP_HOST=127.0.100.1
SP_API_HTTP_PORT=8080
"""

SP_CONF = {
    'SP_API_HTTP_HOST': '127.0.100.1',
    'SP_API_HTTP_PORT': '8080',
    'SP_AUTH_TOKEN': '1234567890123456789',
    'SP_OURID': '1'
}


def faulty_api(req):
    faulty_api.real_fn(req)
    if faulty_api.fail_count > 0:
        faulty_api.fail_count -= 1
        raise storpool_utils.StorPoolAPIError(
            500,
            {
                'error': {
                    'name': 'busy',
                    'descr': "'os--volume--sp-vol-1' is open at client 19"
                }
            })


def _fake_open(path):
    data = ""
    if path.name == '/etc/storpool.conf':
        if STORPOOL_CONF_INI_SELECTOR == 'full':
            data = STORPOOL_CONF_INI
        if STORPOOL_CONF_INI_SELECTOR == 'no-hostname':
            data = STORPOOL_CONF_INI_NO_HOSTNAME
    elif path.name == '/etc/storpool.conf.d/another.conf':
        data = ANOTHER_CONF_INI
    else:
        raise Exception(f"Called open with an unexpected path: {path}")

    open_mock = mock.Mock()
    open_mock.read = lambda: data
    ctx_mock = mock.Mock()
    ctx_mock.__enter__ = mock.Mock(return_value=open_mock)
    ctx_mock.__exit__ = mock.Mock()
    return ctx_mock


def _fake_node():
    return 'this-node'


class FakePath:
    def __init__(self, name, exists, is_file, dir_contents = None):
        self.name = name
        self.exists = exists
        self.is_a_file = is_file
        self.dir_contents = dir_contents

    def is_file(self):
        return self.exists and self.is_a_file

    def is_dir(self):
        return self.exists and not self.is_a_file

    def iterdir(self):
        if self.dir_contents is None:
            raise Exception(
                f"Called iterdir() on a non-directory: {self.name}")
        return self.dir_contents

    def __str__(self):
        return self.name


@mock.patch('builtins.open', _fake_open)
@mock.patch('platform.node', _fake_node)
class StorPoolConfTestCase(base.TestCase):
    def setUp(self):
        super(StorPoolConfTestCase, self).setUp()

        self.mock_path = mock.Mock()
        self.fs_tree = [
            FakePath('/etc/storpool.conf', True, True),
            FakePath('/etc/storpool.conf.d', True, False, [])
        ]

    def test_subconf_overrides_main(self):
        self.fs_tree[1] = FakePath('/etc/storpool.conf.d', True, False, [
            FakePath('/etc/storpool.conf.d/another.conf', True, True)
        ])
        self._fs_init(self.fs_tree, 'full')

        with mock.patch('pathlib.Path', self.mock_path):
            conf = storpool_utils.get_conf()

        self.assertEqual(SP_CONF, conf)

    def test_only_storpool_conf(self):
        self._fs_init(self.fs_tree, 'full')

        sp_conf_expected = copy.deepcopy(SP_CONF)
        sp_conf_expected['SP_API_HTTP_HOST'] = '127.0.0.1'
        sp_conf_expected['SP_API_HTTP_PORT'] = '81'

        with mock.patch('pathlib.Path', self.mock_path):
            conf = storpool_utils.get_conf()

        self.assertEqual(sp_conf_expected, conf)

    def test_env_overrides_main(self):
        self._fs_init(self.fs_tree, 'full')

        overrides_expected = {
            'SP_API_HTTP_HOST': '192.168.0.10',
            'SP_API_HTTP_PORT': '8123'
        }

        sp_conf_expected = copy.deepcopy(SP_CONF)
        sp_conf_expected.update(overrides_expected)

        with (mock.patch('pathlib.Path', self.mock_path),
              mock.patch.dict(os.environ, overrides_expected)):
            conf = storpool_utils.get_conf()

        self.assertEqual(sp_conf_expected, conf)

    def test_raise_if_no_storpool_conf(self):
        self.fs_tree[0] = FakePath('/etc/storpool.conf', False, True)
        self._fs_init(self.fs_tree, 'full')

        with mock.patch('pathlib.Path', self.mock_path):
            self.assertRaises(exception.BrickException,
                              storpool_utils.get_conf)

    def _fs_init(self, fs, storpool_conf_type):
        global STORPOOL_CONF_INI_SELECTOR
        STORPOOL_CONF_INI_SELECTOR = storpool_conf_type
        self.mock_path.side_effect = fs


class StorPoolAPITestCase(base.TestCase):
    def setUp(self):
        super(StorPoolAPITestCase, self).setUp()
        self.api = storpool_utils.StorPoolAPI(
            '127.0.0.1', '81', '1234567890123456789')

    def test_api_ok(self):
        with mock.patch('http.client.HTTPConnection') as connection_mock:
            resp = mock.Mock()
            c_mock = connection_mock.return_value
            c_mock.getresponse = mock.Mock(return_value=resp)

            resp.status = 200
            resp.read = lambda: '{ "data": [{ "name": "test-volume" }] }'

            self.assertEqual(self.api.volumes_list(),
                             [{'name': 'test-volume'}])

    def test_api_exceptions(self):
        with mock.patch('http.client.HTTPConnection') as connection_mock:
            resp = mock.Mock()
            c_mock = connection_mock.return_value
            c_mock.getresponse = mock.Mock(return_value=resp)

            resp.status = 200
            resp.read = lambda: '{}'
            self.assertRaises(KeyError, self.api.volumes_list)

            resp.read = lambda: '{/}'
            self.assertRaises(json.JSONDecodeError, self.api.volumes_list)

            resp.read = lambda: '{ "error": { "transient": true } }'
            self.assertRaises(storpool_utils.StorPoolAPIError,
                              self.api.volumes_list)

    def test_api_handle_transient(self):
        with mock.patch('http.client.HTTPConnection') as connection_mock:
            resp = mock.Mock()
            resp.status = 500
            resp.read = lambda: '{ "error": { "transient": true } }'

            resp1 = mock.Mock()
            resp1.status = 200
            resp1.read = lambda: '{ "data": [{ "name": "test-volume" }] }'

            c_mock = connection_mock.return_value
            c_mock.getresponse = mock.Mock(side_effect=[resp, resp, resp1])

            self.assertEqual(self.api.volumes_list(),
                             [{'name': 'test-volume'}])
